from rich.panel import Panel
from rich.tree import Tree
from typing import Dict
from blok.blok import Blok


def create_welcome_pane(path, blok):
    return Panel(
        f"We will generate your new setup here: {path} using the {blok} blok",
        expand=False,
        title="Welcome to Blok!",
        style="bold magenta",
    )


def create_dependency_resolutions_pane(dep_map: Dict[str, Blok]):
    tree = Tree("Dependency Resolutions")

    for key, value in dep_map.items():
        tree.add(f"[bold]{key}[/bold] resolved to {value.get_blok_name()}")

    return Panel(
        tree,
        expand=False,
        title="Dependency Resolutions",
        style="magenta",
    )



import typing as t
from blok.errors import ProtocolNotCompliantError
import importlib


def check_allowed_module_string(value: str) -> bool:
    for char in value:
        if char not in "abcdefghijklmnopqrstuvwxyz_":
            return False

    return True


def check_protocol_compliance(instance: t.Type, protocol: t.Type) -> bool:
    protocol_methods = {
        name
        for name in dir(protocol)
        if getattr(protocol, name, None) and not name.startswith("_")
    }

    instance_methods = {
        name
        for name in dir(instance)
        if getattr(instance, name, None) and not name.startswith("_")
    }
    missing_methods = protocol_methods - instance_methods
    if missing_methods:
        raise ProtocolNotCompliantError(
            f"Instance of {type(instance).__name__} does not implement "
            + f"the following methods required by {protocol.__name__}"
            + ":\n\t-"
            + "\n\t -".join(missing_methods)
        )

    return True


def check_service_compliance(cls: t.Type, service: t.Type) -> bool:
    protocol_methods = {
        name
        for name in dir(service)
        if getattr(service, name, None) and not name.startswith("_")
    }

    instance_methods = {
        name
        for name in dir(cls)
        if getattr(cls, name, None) and not name.startswith("_")
    }
    missing_methods = protocol_methods - instance_methods
    if missing_methods:
        raise ProtocolNotCompliantError(
            f"Class {cls.__name__} does not implement "
            + f"the following methods required by {service.__name__}"
            + ":\n\t-"
            + "\n\t -".join(missing_methods)
        )

    return True


def remove_empty_dicts(d):
    if not isinstance(d, dict):
        return d

    non_empty_items = {}
    for key, value in d.items():
        if isinstance(value, dict):
            cleaned_dict = remove_empty_dicts(value)
            if cleaned_dict:  # Only add non-empty dictionaries
                non_empty_items[key] = cleaned_dict
        else:
            non_empty_items[key] = value

    return non_empty_items


def get_prepended_values(kwargs: t.Dict[str, t.Any], blok_name: str):
    prepended = {
        key.split("_", 1)[1]: value
        for key, value in kwargs.items()
        if key.startswith(blok_name)
    }

    return prepended


def get_cleartext_deps(blok):
    dependencies = blok.get_dependencies()
    cleartext_deps = []

    for dep in dependencies:
        if hasattr(dep, "get_identifier"):
            print("Dep is a service/blok class", dep)
            cleartext_deps.append(dep.get_identifier())
        else:
            assert isinstance(
                dep, str
            ), "Dependency must be a string or a service/blok class"
            cleartext_deps.append(dep)

    return cleartext_deps



"""This module contains functionality for comparing
two docker-compose setups


"""

from collections.abc import Iterable
from typing import Protocol, List
from dataclasses import dataclass


class Diffable(Protocol):
    def diff(self, other: "Diffable", path: str) -> List[str]: ...


@dataclass
class ChangedDiff:
    path: str
    old: object
    new: object


@dataclass
class AddedKeyDiff:
    path: str
    new: str


@dataclass
class RemoveKeyDiff:
    path: str
    old: str


@dataclass
class ListChangedDiff:
    path: str
    old: str


Diff = ChangedDiff | AddedKeyDiff | RemoveKeyDiff | ListChangedDiff


def compare_structures(old_structure, new_structure, path=""):
    diffs = []

    old_keys = set(old_structure.keys())
    new_keys = set(new_structure.keys())

    for key in old_keys - new_keys:
        diffs.append(RemoveKeyDiff(path, key))

    for key in new_keys - old_keys:
        diffs.append(AddedKeyDiff(path, key))

    for key in old_keys & new_keys:
        old_value = old_structure[key]
        new_value = new_structure[key]

        if hasattr(old_value, "diff") and hasattr(new_value, "diff"):
            diffs.extend(old_value.diff(new_value, path + "/" + key))

        elif isinstance(old_value, dict) and isinstance(new_value, dict):
            diffs.extend(compare_structures(old_value, new_value, path + "/" + key))

        elif isinstance(old_value, (list, tuple)) and isinstance(new_value, (list, tuple)):
            if len(old_value) != len(new_value):
                diffs.append(
                    ListChangedDiff(
                        path + "/" + key, f"Length mismatch: {len(old_value)} != {len(new_value)}"
                    )
                )

            for i, (a, b) in enumerate(zip(old_value, new_value)):
                diffs.extend(
                    compare_structures(
                        {"a": a},
                        {"a": b},
                        path + "/" + key + "/" + str(i),
                    )
                )

        elif old_value != new_value:
            diffs.append(ChangedDiff(path + "/" + key, old_value, new_value))

    return diffs



from dataclasses import asdict
from pathlib import Path
import traceback
from rich import get_console
import yaml
from blok.diff import compare_structures
from blok.io.read import create_structure_from_files_and_folders
from blok.io.write import create_files_and_folders
from blok.registry import BlokRegistry
from blok.errors import (
    DependencyNotFoundError,
    TooManyBlokFoundError,
    BlokBuildError,
    BlokInitializationError,
)
from blok.tree.models import YamlFile
from blok.utils import get_cleartext_deps, get_prepended_values, remove_empty_dicts
from blok.models import NestedDict
from blok.blok import Command, ExecutionContext, InitContext
import rich_click as click
from rich.tree import Tree
from blok.render.tree import construct_file_tree, construct_diff_tree
from blok.render.panel import create_welcome_pane, create_dependency_resolutions_pane
import os
from blok.blok import Blok
from typing import List, Optional
from collections import OrderedDict
from blok.renderer import Renderer
import subprocess


def filter_bloks(
    bloks: List[Blok],
    discard_bloks: Optional[List[str]] = None,
    prefer_bloks: Optional[List[str]] = None,
):
    if discard_bloks:
        bloks = [blok for blok in list if blok.get_blok_name() not in discard_bloks]

    if prefer_bloks:
        prefered_blok = [blok for blok in list if blok.get_blok_name() in prefer_bloks]
        if prefered_blok:
            bloks = prefered_blok

    return bloks


def traverse_command_tree(nested: NestedDict):
    for key, value in nested.items():
        if isinstance(value, NestedDict):
            yield from traverse_command_tree(value)
        else:
            assert isinstance(value, Command), f"Expected Command but got {value}"
            yield key, asdict(value)


def initialize_blok_with_dependencies(
    blok: Blok,
    registry: BlokRegistry,
    discard_bloks: list,
    prefer_bloks: list,
    kwargs: dict,
):
    """
    Given a blok name, initialize the necessary modules including its dependencies.

    :param blok_name: The name of the blok to initialize.
    """
    initialied_bloks = OrderedDict()

    def initialize_service_as_blok_recursive(service_name, causing_blok):
        if service_name in initialied_bloks:
            return

        try:
            potential_bloks = registry.get_bloks_for_dependency(service_name)
        except KeyError:
            raise DependencyNotFoundError(
                f"Could not find blok for key {service_name}. This might happen if you have not registered a blok fullfilling the interface Causing Blok: {causing_blok}"
            )

        filtered_bloks = filter_bloks(
            potential_bloks, discard_bloks=discard_bloks, prefer_bloks=prefer_bloks
        )

        if not filtered_bloks:
            raise DependencyNotFoundError(
                f"Could not find blok for key {service_name}. This might happen if you have not registered a blok fullfilling the interface"
            )

        if len(filtered_bloks) > 1:
            raise TooManyBlokFoundError(
                f"Multiple bloks found for key {service_name}. This might happen if you have registered multiple bloks fullfilling the same interface. Just discard the ones you do not need. Or set your prefered bloks. Here are the bloks found: {filtered_bloks}"
            )

        chosen_blok = filtered_bloks[0]

        initialied_bloks[service_name] = chosen_blok

        for dependency in get_cleartext_deps(chosen_blok):
            initialize_service_as_blok_recursive(dependency, chosen_blok)

        blok_dependencies = {
            key: initialied_bloks[key] for key in get_cleartext_deps(chosen_blok)
        }

        try:
            chosen_blok.preflight(
                InitContext(
                    dependencies=blok_dependencies,
                    kwargs=get_prepended_values(kwargs, chosen_blok.get_blok_name()),
                )
            )
        except Exception as e:
            raise BlokInitializationError(
                f"Failed to initialize blok {chosen_blok.get_blok_name()}"
            ) from e

    for blok_key in get_cleartext_deps(blok):
        initialize_service_as_blok_recursive(blok_key, blok)

    blok_dependencies = {key: initialied_bloks[key] for key in get_cleartext_deps(blok)}

    try:
        blok.preflight(
            InitContext(
                dependencies=blok_dependencies,
                kwargs=get_prepended_values(kwargs, blok.get_blok_name()),
            )
        )
    except Exception as e:
        raise BlokInitializationError(
            f"Failed to initialize starting blok {blok.get_blok_name()}"
        ) from e

    return initialied_bloks


def entrypoint(
    registry: BlokRegistry, renderer: Renderer, blok_file_name: str, **kwargs
):
    try:
        console = get_console()
        path = kwargs.pop("path")
        blok_name = kwargs.pop("blok")
        yes = kwargs.pop("yes", False)

        dry = kwargs.pop("dry", False)

        discard_bloks = kwargs.pop("discard_bloks", None)
        prefer_bloks = kwargs.pop("prefer_bloks", None)

        blok = registry.get_blok(blok_name)
        blok.entry(renderer)

        initialized = initialize_blok_with_dependencies(
            blok, registry, discard_bloks, prefer_bloks, kwargs
        )

        pane = create_dependency_resolutions_pane(initialized)
        console.print(pane)

        files = {}

        context = ExecutionContext(
            docker_compose=NestedDict({"services": {}, "networks": {}}),
            file_tree=NestedDict(files),
            install_commands=NestedDict(),
            up_commands=NestedDict(),
        )

        for key, service in initialized.items():
            try:
                service.build(context)
            except Exception as e:
                raise BlokInitializationError(
                    f"Failed to initialize blok {blok.get_blok_name()} for service {key}"
                ) from e

            # TODO: Validate context?

        # Finally build the blok
        try:
            blok.build(context)
        except Exception as e:
            raise BlokInitializationError(
                f"Failed to initialize blok {blok.get_blok_name()} for service {key}"
            ) from e

        # This would generate this are you okay?

        compose_dict = remove_empty_dicts(context.docker_compose)

        print_all = False

        if print_all:
            tree = construct_file_tree("DockerCompose", compose_dict)
            get_console().print(tree)

            tree = construct_file_tree("Files", context.file_tree)
            get_console().print(tree)

        old_files = create_structure_from_files_and_folders(path)

        new_yaml_body = {
            "config": kwargs,
            "resolved_blocks": {
                key: value.get_blok_name() for key, value in initialized.items()
            },
            "initialized_order": list(i.get_blok_name() for i in initialized.values()),
            "install_commands": dict(traverse_command_tree(context.install_commands)),
            "up_commands": dict(traverse_command_tree(context.up_commands)),
        }

        diffs = compare_structures(
            old_files,
            {
                **context.file_tree,
                "docker-compose.yml": YamlFile(**compose_dict),
                "__blok__.yml": YamlFile(**new_yaml_body),
            },
        )

        if not diffs:
            get_console().print("No differences found")
            get_console().print("Exiting")

        else:
            tree = construct_diff_tree(diffs)
            get_console().print(tree)

        # Generate docker compose file

        if yes or renderer.confirm("Do you want to generate the files?"):
            os.makedirs(path, exist_ok=True)

            docker_compose_file_path = Path(path) / "docker-compose.yml"

            with open(docker_compose_file_path, "w") as f:
                yaml.dump(compose_dict, f, yaml.SafeDumper, indent=3)

            print(f"Generated docker-compose file at {docker_compose_file_path}")

            # Generate file tree

            # Generate files and folders
            create_files_and_folders(Path(path), context.file_tree)

        with open(Path(path) / blok_file_name, "w") as f:
            yaml.dump(
                new_yaml_body,
                f,
                yaml.SafeDumper,
            )

    except Exception as e:
        print(traceback.format_exc())
        raise click.ClickException(str(e)) from e



from pathlib import Path
import signal
import time
import subprocess
import click
from blok.blok import Command
from blok.registry import BlokRegistry
from blok.renderer import Renderer, Panel

# List to keep track of subprocesses
subprocesses = []


def signal_handler(sig, frame):
    print("Main process received interrupt signal")
    terminate_all_subprocesses()
    print("All subprocesses terminated")
    exit(0)


def terminate_all_subprocesses():
    for p in subprocesses:
        if p.poll() is None:  # Check if process is still running
            print(f"Terminating subprocess with PID: {p.pid}")
            p.terminate()
    for p in subprocesses:
        p.wait()


def secure_path_combine(x: Path, y: Path) -> Path:
    # Resolve the combined path
    mother_path = x.resolve()
    combined_path = (mother_path / y).resolve()

    # Check if the combined path is within the mother path
    if mother_path in combined_path.parents or mother_path == combined_path:
        return combined_path
    else:
        raise ValueError(
            f"The user-defined path traverses out of the mother path. {list(combined_path.parents)} but requested {mother_path}"
        )


@click.pass_context
def entrypoint(
    ctx: click.Context,
    registry: BlokRegistry,
    renderer: Renderer,
    blok_file_name: str,
    **kwargs,
):
    renderer.render(
        Panel("Lets up this project", title="Welcome to Blok!", style="bold magenta")
    )

    path = Path(kwargs.pop("path"))
    yes = kwargs.pop("yes", False)
    select = kwargs.pop("select", [])

    up_commands = ctx.obj.get("up_commands", {})

    selected_commands = []

    if select:
        print("Selecting commands")
        for key in select:
            if key in up_commands:
                selected_commands.append(up_commands[key])
            else:
                raise click.ClickException(f"Command with key {key} not found")

    else:
        selected_commands = list(up_commands.values())
        selected_commands.append(
            {"command": ["docker", "compose", "up", "-d"], "cwd": "."}
        )

    if yes or renderer.confirm(
        "Run up commands? \n\t"
        + "\n\t".join(map(lambda x: " ".join(x["command"]), selected_commands))
        + "\n"
    ):
        print("Running up commands")
    else:
        raise click.Abort("User aborted")
        return

    if len(selected_commands) == 1:
        command = selected_commands[0]
        rel_path = secure_path_combine(path, Path(command["cwd"]))
        p = subprocess.run(
            " ".join(command["command"]),
            shell=True,
            cwd=rel_path,
        )

    else:
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        for command in selected_commands:
            rel_path = secure_path_combine(path, Path(command["cwd"]))
            p = subprocess.Popen(
                command["command"],
                cwd=rel_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            subprocesses.append(p)
            print(f"Started subprocess {command['command']} with PID: {p.pid}")

        try:
            while True:
                for p in subprocesses:
                    if p.poll() is not None:  # Process has terminated
                        if p.returncode != 0:  # Process failed
                            error_output = p.stderr.read().decode()
                            print(
                                f"Subprocess with PID {p.pid} failed with return code {p.returncode}"
                            )
                            print(f"Error output: {error_output}")
                            terminate_all_subprocesses()
                            raise click.ClickException(
                                f"Subprocess failed: {error_output}"
                            )
                time.sleep(1)
        except KeyboardInterrupt:
            signal_handler(signal.SIGINT, None)



from typing import List, runtime_checkable, Protocol
import typing as t


class Service(Protocol):
    def get_identifier(self) -> str:
        raise NotImplementedError("This method must be implemented by the subclass")


def service(name: str):
    def decorator(cls):
        if not hasattr(cls, "get_identifier"):
            cls.get_identifier = classmethod(lambda self: name)

        try:
            return runtime_checkable(cls)
        except TypeError as e:
            raise TypeError(f"Could not create Blok Service from {cls}. Blok services need to inherit from 'typing.Protocol'") from e

    return decorator
